import pytest

from pathlib import Path
from math import isclose

import micrometa.micrometa as meta

data_dir = Path("tests", "data")
metadata_dir = data_dir / "metadata"

cellfinder_metadata = metadata_dir / "cellfinder_metadata.ini"
baking_tray_metadata = metadata_dir / "BakingTray_recipe.yml"
mesoSPIM_metadata = metadata_dir / "mesoSPIM.raw_meta.txt"
unsupported_metadata = metadata_dir / "unsupported_metadata.txt"
missing_metadata = metadata_dir / "cellfinder_metadata_missing.ini"

VOX_DIM_TOLERANCE = 0.1


def test_cellfinder_meta():
    metadata = meta.get_acqusition_metadata(cellfinder_metadata)
    assert isclose(2, metadata.x_pixel_um, abs_tol=VOX_DIM_TOLERANCE)
    assert isclose(2, metadata.y_pixel_um, abs_tol=VOX_DIM_TOLERANCE)
    assert isclose(5, metadata.z_pixel_um, abs_tol=VOX_DIM_TOLERANCE)

    assert metadata.num_planes == 1500


def test_baking_tray_meta():
    metadata = meta.get_acqusition_metadata(baking_tray_metadata)
    assert isclose(2.19, metadata.x_pixel_um, abs_tol=VOX_DIM_TOLERANCE)
    assert isclose(2.14, metadata.y_pixel_um, abs_tol=VOX_DIM_TOLERANCE)
    assert isclose(5, metadata.z_pixel_um, abs_tol=VOX_DIM_TOLERANCE)

    assert metadata.num_planes == 2000


def test_mesospim_meta():
    metadata = meta.get_acqusition_metadata(mesoSPIM_metadata)
    assert isclose(8.23, metadata.x_pixel_um, abs_tol=VOX_DIM_TOLERANCE)
    assert isclose(8.23, metadata.y_pixel_um, abs_tol=VOX_DIM_TOLERANCE)
    assert isclose(10, metadata.z_pixel_um, abs_tol=VOX_DIM_TOLERANCE)

    assert metadata.num_planes == 852


def test_unsupported_meta():
    with pytest.raises(NotImplementedError):
        assert meta.get_acqusition_metadata(unsupported_metadata)


class Args:
    def __init__(self):
        self.x_pixel_um = self.y_pixel_um = self.z_pixel_um = None
        self.metadata = None

    def set_all_pixel_sizes(self):
        self.x_pixel_um = 1
        self.y_pixel_um = 2
        self.z_pixel_um = 3

    def set_some_pixel_sizes_w_meta_baking_tray(self):
        self.x_pixel_um = 10
        self.y_pixel_um = None
        self.z_pixel_um = 40

        self.metadata = baking_tray_metadata

    def set_some_pixel_sizes_w_meta_mesospim(self):
        self.x_pixel_um = 100
        self.y_pixel_um = None
        self.z_pixel_um = None

        self.metadata = mesoSPIM_metadata

    def set_some_pixel_sizes_w_meta_cellfinder(self):
        self.x_pixel_um = None
        self.y_pixel_um = 0.2
        self.z_pixel_um = 40

        self.metadata = cellfinder_metadata

    def set_some_pixel_sizes_no_meta(self):
        self.x_pixel_um = None
        self.y_pixel_um = 100
        self.z_pixel_um = 3
        self.metadata = None

    def set_some_pixel_sizes_unsupported_meta(self):
        self.x_pixel_um = None
        self.y_pixel_um = 100
        self.z_pixel_um = 3
        self.metadata = unsupported_metadata

    def set_some_pixel_sizes_missing_meta(self):
        self.x_pixel_um = None
        self.y_pixel_um = 100
        self.z_pixel_um = 3
        self.metadata = missing_metadata



import pytest

from random import randint
from argparse import ArgumentTypeError
from pathlib import Path

from micrometa import tools

data_dir = Path("tests", "data")
jabberwocky = data_dir / "general" / "jabberwocky.txt"


def test_check_positive_float():
    pos_val = randint(1, 1000) / 100
    neg_val = -randint(1, 1000) / 100

    assert pos_val == tools.check_positive_float(pos_val)

    with pytest.raises(ArgumentTypeError):
        assert tools.check_positive_float(neg_val)

    assert tools.check_positive_float(None) is None

    with pytest.raises(ArgumentTypeError):
        assert tools.check_positive_float(None, none_allowed=False)

    assert tools.check_positive_float(0) == 0


def test_check_positive_int():
    pos_val = randint(1, 1000)
    neg_val = -randint(1, 1000)

    assert pos_val == tools.check_positive_int(pos_val)

    with pytest.raises(ArgumentTypeError):
        assert tools.check_positive_int(neg_val)

    assert tools.check_positive_int(None) is None

    with pytest.raises(ArgumentTypeError):
        assert tools.check_positive_int(None, none_allowed=False)

    assert tools.check_positive_int(0) == 0


def test_get_text_lines():
    line_5 = "The jaws that bite, the claws that catch!"
    first_line_alphabetically = "All mimsy were the borogoves,"
    assert tools.get_text_lines(jabberwocky, return_lines=5) == line_5
    assert (
        tools.get_text_lines(
            jabberwocky, return_lines=6, remove_empty_lines=False
        )
        == line_5
    )
    assert (
        tools.get_text_lines(jabberwocky, sort=True)[0]
        == first_line_alphabetically
    )



"""
micrometa
===============

Functions to read image acqusition metadata files.

"""

import configparser
import yaml

from . import tools

SUPPORTED_METADATA_TYPES = ["Cellfinder", "BakingTray", "mesoSPIM"]

CELLFINDER_STR = "cellfinder_meta"
CELLFINDER_SUFFIX = ".ini"

BAKING_TRAY_STR = "recipe"
BAKING_TRAY_SUFFIX = ".yml"

MESOSPIM_STR = "raw_meta"
MESOSPIM_SUFFIX = ".txt"


def get_acqusition_metadata(metadata_path):
    """
    Parses metadata files. Will guess what kind of metadata it is.
    :param metadata_path: Pathlib object referencing the metadata file
    :return: Object with potentially useful metadata attributes
    """
    if (BAKING_TRAY_STR in metadata_path.name) and (
        metadata_path.suffix == BAKING_TRAY_SUFFIX
    ):
        metadata = get_baking_tray_metadata(metadata_path)

    elif (MESOSPIM_STR in metadata_path.name) and (
        metadata_path.suffix == MESOSPIM_SUFFIX
    ):
        metadata = get_mesospim_metadata(metadata_path)

    elif CELLFINDER_STR in metadata_path.name and (
        metadata_path.suffix == CELLFINDER_SUFFIX
    ):
        metadata = get_cellfinder_metadata(metadata_path)

    else:
        raise NotImplementedError(
            "The metadata type of file: '{}' is not yet supported. Please "
            "supply of one of: '{}' or enter metadata manually.".format(
                metadata_path.name, SUPPORTED_METADATA_TYPES
            )
        )

    return metadata


class CellfinderMetadata:
    """
    Class to deal with cellfinder custom metadata files
    """

    def __init__(self, metadata_path):
        self._metadata_path = metadata_path
        self._metadata = read_ini(metadata_path)

    @property
    def x_pixel_um(self):
        pixel_um = self._metadata["voxels"].getfloat("x")
        tools.check_positive_float(pixel_um, none_allowed=False)
        return pixel_um

    @property
    def y_pixel_um(self):
        pixel_um = self._metadata["voxels"].getfloat("y")
        tools.check_positive_float(pixel_um, none_allowed=False)
        return pixel_um

    @property
    def z_pixel_um(self):
        pixel_um = self._metadata["voxels"].getfloat("z")
        tools.check_positive_float(pixel_um, none_allowed=False)
        return pixel_um

    @property
    def num_planes(self):
        num_planes = self._metadata["planes"].getint("num_planes")
        tools.check_positive_int(num_planes, none_allowed=False)
        return num_planes


class BakingTrayMetadata:
    """
    Class to deal with BakingTray
    (https://github.com/SainsburyWellcomeCentre/BakingTray) recipe files:
    """

    def __init__(self, metadata_path):
        self._metadata_path = metadata_path
        self._metadata = read_yaml(metadata_path)

    @property
    def x_pixel_um(self):
        pixel_um = float(
            self._metadata["StitchingParameters"]["VoxelSize"]["X"]
        )
        tools.check_positive_float(pixel_um, none_allowed=False)
        return pixel_um

    @property
    def y_pixel_um(self):
        pixel_um = float(
            self._metadata["StitchingParameters"]["VoxelSize"]["Y"]
        )
        tools.check_positive_float(pixel_um, none_allowed=False)
        return pixel_um

    @property
    def z_pixel_um(self):
        slice_thickness = float(self._metadata["mosaic"]["sliceThickness"])
        num_optical_planes = float(
            self._metadata["mosaic"]["numOpticalPlanes"]
        )
        pixel_um = 1000 * slice_thickness / num_optical_planes
        tools.check_positive_float(pixel_um, none_allowed=False)
        return pixel_um

    @property
    def num_planes(self):
        num_physical_sections = float(self._metadata["mosaic"]["numSections"])
        num_optical_planes = float(
            self._metadata["mosaic"]["numOpticalPlanes"]
        )
        num_planes = num_physical_sections * num_optical_planes
        tools.check_positive_int(num_planes, none_allowed=False)
        return num_planes


class MesoSpimMetadata:
    """
    Class to deal with mesoSPIM (https://github.com/mesoSPIM/mesoSPIM-control)
    metadata files:
    """

    def __init__(self, metadata_path):
        self._metadata_path = metadata_path
        self._metadata = tools.get_text_lines(metadata_path)

    @property
    def x_pixel_um(self):
        line = [
            item
            for item in self._metadata
            if item.startswith("[Pixelsize in um]")
        ][0]
        pixel_um = float(line.split()[-1])
        tools.check_positive_float(pixel_um, none_allowed=False)
        return pixel_um

    @property
    def y_pixel_um(self):
        line = [
            item
            for item in self._metadata
            if item.startswith("[Pixelsize in um]")
        ][0]
        pixel_um = float(line.split()[-1])
        tools.check_positive_float(pixel_um, none_allowed=False)
        return pixel_um

    @property
    def z_pixel_um(self):
        line = [
            item for item in self._metadata if item.startswith("[z_stepsize]")
        ][0]
        pixel_um = float(line.split()[-1])
        tools.check_positive_float(pixel_um, none_allowed=False)
        return pixel_um

    @property
    def num_planes(self):
        line = [
            item for item in self._metadata if item.startswith("[z_planes]")
        ][0]
        num_planes = int(line.split()[-1])
        tools.check_positive_int(num_planes, none_allowed=False)
        return num_planes


def get_cellfinder_metadata(metadata_path):
    """
    Parses cellfinder custom metadata files.
    :param metadata_path: Pathlib object referencing the metadata file
    :return: Object with potentially useful metadata attributes
    """

    return CellfinderMetadata(metadata_path)


def get_baking_tray_metadata(metadata_path):
    """
    Parses BakingTray (https://github.com/SainsburyWellcomeCentre/BakingTray)
    recipe files.
    :param metadata_path: Pathlib object referencing the metadata file
    :return: Object with potentially useful metadata attributes
    """

    return BakingTrayMetadata(metadata_path)


def get_mesospim_metadata(metadata_path):
    """
    Parses mesoSPIM (https://github.com/mesoSPIM/mesoSPIM-control)
    metadata files.
    :param metadata_path: Pathlib object referencing the metadata file
    :return: Object with potentially useful metadata attributes
    """

    return MesoSpimMetadata(metadata_path)


def read_ini(ini_path):
    """
    Returns a configparser object to be parsed.
    :param ini_path: Path to ini file
    :return: configparser object to be parsed.
    """

    config = configparser.ConfigParser()
    config.read(ini_path)
    return config


def read_yaml(yaml_path):
    """
    Returns a yaml object to be parsed.
    :param yaml_path: Path to yaml file
    :return: yaml object to be parsed.
    """

    with open(yaml_path, "r") as stream:
        return yaml.safe_load(stream)



import argparse
from natsort import natsorted


def check_positive_float(value, none_allowed=True):
    """
    Used in argparse to enforce positive floats
    FromL https://stackoverflow.com/questions/14117415
    :param value: Input value
    :param none_allowed: If false, throw an error for None values
    :return: Input value, if it's positive
    """
    ivalue = value
    if ivalue is not None:
        ivalue = float(ivalue)
        if ivalue < 0:
            raise argparse.ArgumentTypeError(
                "%s is an invalid positive value" % value
            )
    else:
        if not none_allowed:
            raise argparse.ArgumentTypeError("%s is an invalid value." % value)

    return ivalue


def check_positive_int(value, none_allowed=True):
    """
    Used in argparse to enforce positive ints
    FromL https://stackoverflow.com/questions/14117415
    :param value: Input value
    :param none_allowed: If false, throw an error for None values
    :return: Input value, if it's positive
    """
    ivalue = value
    if ivalue is not None:
        ivalue = int(ivalue)
        if ivalue < 0:
            raise argparse.ArgumentTypeError(
                "%s is an invalid positive value" % value
            )
    else:
        if not none_allowed:
            raise argparse.ArgumentTypeError("%s is an invalid value." % value)

    return ivalue


def remove_empty_string_list(str_list):
    """
    Removes any empty strings from a list of strings
    :param str_list: List of strings
    :return: List of strings without the empty strings
    """
    return list(filter(None, str_list))


def get_text_lines(
    file, return_lines=None, rstrip=True, sort=False, remove_empty_lines=True
):
    """
    Return only the nth line of a text file
    :param file: Any text file
    :param return_lines: Which specific line/lines to read
    :param rstrip: Remove trailing characters
    :param sort: If true, naturally sort the data
    :param remove_empty_lines: If True, ignore empty lines
    :return: The nth line
    """
    with open(file) as f:
        lines = f.readlines()
    if rstrip:
        lines = [line.strip() for line in lines]
    if remove_empty_lines:
        lines = remove_empty_string_list(lines)
    if sort:
        lines = natsorted(lines)
    if return_lines is not None:
        lines = lines[return_lines]
    return lines



from setuptools import setup, find_packages

requirements = ["configparser", "natsort", "pyyaml"]

setup(
    name="micrometa",
    version="0.0.3",
    description="Reading of microsocopy metadata",
    install_requires=requirements,
    extras_require={"dev": ["black", "pytest-cov", "pytest"]},
    python_requires=">=3.6",
    packages=find_packages(),
    include_package_data=True,
    url="https://github.com/adamltyson/micrometa",
    author="Adam Tyson",
    author_email="adam.tyson@ucl.ac.uk",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
    zip_safe=False,
)
